/*
 *  Copyright (c) 2013 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

// Unit tests for PostDecodeVad class.

#include "webrtc/modules/audio_coding/neteq/post_decode_vad.h"

#include "testing/gtest/include/gtest/gtest.h"

namespace webrtc {

TEST(PostDecodeVad, CreateAndDestroy) {
  PostDecodeVad vad;
}

// TODO(hlundin): Write more tests.

}  // namespace webrtc
