/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

/**
 * Site Data - Settings dialog
 */
#sitesList {
  min-height: 20em;
}

.item-box {
  padding: 5px 8px;
}

/**
 * Confirmation dialog of removing sites selected
 */
#SiteDataRemoveSelectedDialog {
  padding: 16px;
}

#contentContainer {
  font-size: 1.2em;
  margin-bottom: 10px;
}

.question-icon {
  margin: 6px;
}

#removing-label {
  font-weight: bold;
}

#sitesTree {
  height: 15em;
}
